#    This file is part of Metasm, the Ruby assembly manipulation suite
#    Copyright (C) 2007 Yoann GUILLOT
#
#    Licence is LGPL, see LICENCE in the top-level directory


require 'metasm/ppc/main'

module Metasm
class PowerPC
	def init
		map = <<EOL.split("\n").map { |l| l[0, 30] }
X   31 375     36 lhaux        Load Halfword Algebraic with Update Indexed
X   31 343     36 lhax         Load Halfword Algebraic Indexed
X   31 790     44 lhbrx        Load Halfword Byte-Reverse Indexed
D   40         35 lhz          Load Halfword and Zero
D   41         35 lhzu         Load Halfword and Zero with Update
X   31 311     35 lhzux        Load Halfword and Zero with Update Indexed
X   31 279     35 lhzx         Load Halfword and Zero Indexed
D   46         46 lmw          Load Multiple Word
X   31 597     48 lswi         Load String Word Immediate
X   31 533     48 lswx         Load String Word Indexed
DS  58   2     38 lwa          Load Word Algebraic
X   31  20     II lwarx        Load Word And Reserve Indexed
X   31 373     38 lwaux        Load Word Algebraic with Update Indexed
X   31 341     38 lwax         Load Word Algebraic Indexed
X   31 534     44 lwbrx        Load Word Byte-Reverse Indexed
D   32         37 lwz          Load Word and Zero
D   33         37 lwzu         Load Word and Zero with Update
X   31  55     37 lwzux        Load Word and Zero with Update Indexed
X   31  23     37 lwzx         Load Word and Zero Indexed
XL  19   0     30 mcrf         Move Condition Register Field
X   63  64    120 mcrfs        Move to Condition Register from FPSCR
X   31 512    135 mcrxr        Move to Condition Register from XER
XFX 31  19     83 mfcr         Move From Condition Register
XFX 31  19    124 mfocrf       Move From One Condition Register Field
X   63 583    120 mffs[.]      Move From FPSCR
X   31  83    III mfmsr        Move From Machine State Register
XFX 31 339     82 mfspr        Move From Special Purpose Register
X   31 595 32 III mfsr         Move From Segment Register
X   31 659 32 III mfsrin       Move From Segment Register Indirect
XFX 31 371     II mftb         Move From Time Base
XFX 31 144     83 mtcrf        Move To Condition Register Fields
XFX 31 144    124 mtocrf       Move To One Condition Register Field
X   63  70    122 mtfsb0[.]    Move To FPSCR Bit 0
X   63  38    122 mtfsb1[.]    Move To FPSCR Bit 1
XFL 63 711    121 mtfsf[.]     Move To FPSCR Fields
X   63 134    121 mtfsfi[.]    Move To FPSCR Field Immediate
X   31 146    III mtmsr        Move To Machine State Register
X   31 178    III mtmsrd       Move To Machine State Register Doubleword
XFX 31 467     81 mtspr        Move To Special Purpose Register
X   31 210 32 III mtsr         Move To Segment Register
X   31 242 32 III mtsrin       Move To Segment Register Indirect
XO  31  73 SR  57 mulhd[.]     Multiply High Doubleword
XO  31   9 SR  57 mulhdu[.]    Multiply High Doubleword Unsigned
XO  31  75 SR  57 mulhw[.]     Multiply High Word
XO  31  11 SR  57 mulhwu[.]    Multiply High Word Unsigned
XO  31 233 SR  56 mulld[o][.]  Multiply Low Doubleword
D    7         56 mulli        Multiply Low Immediate
XO  31 235 SR  56 mullw[o][.]  Multiply Low Word
X   31 476 SR  67 nand[.]      NAND
XO  31 104 SR  55 neg[o][.]    Negate
X   31 124 SR  68 nor[.]       NOR
X   31 444 SR  67 or[.]        OR
X   31 412 SR  68 orc[.]       OR with Complement
D   24         66 ori          OR Immediate
D   25         66 oris         OR Immediate Shifted
X   31 122    70  popcntb      Population Count Bytes
XL  19  18    III rfid         Return from Interrupt Doubleword
MDS 30   8 SR  74 rldcl[.]     Rotate Left Doubleword then Clear Left
MDS 30   9 SR  75 rldcr[.]     Rotate Left Doubleword then Clear Right
MD  30   2 SR  73 rldic[.]     Rotate Left Doubleword Immediate then Clear
MD  30   0 SR  72 rldicl[.]    Rotate Left Doubleword Immediate then Clear Left
MD  30   1 SR  72 rldicr[.]    Rotate Left Doubleword Immediate then Clear Right
MD  30   3 SR  76 rldimi[.]    Rotate Left Doubleword Immediate then Mask Insert
M   20     SR  76 rlwimi[.]    Rotate Left Word Immediate then Mask Insert
M   21     SR  73 rlwinm[.]    Rotate Left Word Immediate then AND with Mask
M   23     SR  75 rlwnm[.]     Rotate Left Word then AND with Mask
SC  17         26 sc           System Call
X   31 498    III slbia        SLB Invalidate All
X   31 434    III slbie        SLB Invalidate Entry
X   31 915    III slbmfee      SLB Move From Entry ESID
X   31 851    III slbmfev      SLB Move From Entry VSID
X   31 402    III slbmte       SLB Move To Entry
X   31  27 SR  77 sld[.]       Shift Left Doubleword
X   31  24 SR  77 slw[.]       Shift Left Word
X   31 794 SR  80 srad[.]      Shift Right Algebraic Doubleword
XS  31 413 SR  79 sradi[.]     Shift Right Algebraic Doubleword Immediate
X   31 792 SR  80 sraw[.]      Shift Right Algebraic Word
X   31 824 SR  79 srawi[.]     Shift Right Algebraic Word Immediate
X   31 539 SR  78 srd[.]       Shift Right Doubleword
X   31 536 SR  78 srw[.]       Shift Right Word
D   38         40 stb          Store Byte
D   39         40 stbu         Store Byte with Update
X   31 247     40 stbux        Store Byte with Update Indexed
X   31 215     40 stbx         Store Byte Indexed
DS  62  0      43 std          Store Doubleword
X   31 214     II stdcx.       Store Doubleword Conditional Indexed
DS  62  1      43 stdu         Store Doubleword with Update
X   31 181     43 stdux        Store Doubleword with Update Indexed
X   31 149     43 stdx         Store Doubleword Indexed
D   54        108 stfd         Store Floating-Point Double
D   55        108 stfdu        Store Floating-Point Double with Update
X   31 759    108 stfdux       Store Floating-Point Double with Update Indexed
X   31 727    108 stfdx        Store Floating-Point Double Indexed
X   31 983    109 stfiwx       Store Floating-Point as Integer Word Indexed
D   52        107 stfs         Store Floating-Point Single
D   53        107 stfsu        Store Floating-Point Single with Update
X   31 695    107 stfsux       Store Floating-Point Single with Update Indexed
X   31 663    107 stfsx        Store Floating-Point Single Indexed
D   44         41 sth          Store Halfword
X   31 918     44 sthbrx       Store Halfword Byte-Reverse Indexed
D   45         41 sthu         Store Halfword with Update
X   31 439     41 sthux        Store Halfword with Update Indexed
X   31 407     41 sthx         Store Halfword Indexed
D   47         46 stmw         Store Multiple Word
X   31 725     49 stswi        Store String Word Immediate
X   31 661     49 stswx        Store String Word Indexed
D   36         42 stw          Store Word
X   31 662     45 stwbrx       Store Word Byte-Reverse Indexed
X   31 150     II stwcx.       Store Word Conditional Indexed
D   37         42 stwu         Store Word with Update
X   31 183     42 stwux        Store Word with Update Indexed
X   31 151     42 stwx         Store Word Indexed
XO  31  40 SR  52 subf[o][.]   Subtract From
XO  31  8  SR  53 subfc[o][.]  Subtract From Carrying
XO  31 136 SR  54 subfe[o][.]  Subtract From Extended
D    8     SR  53 subfic       Subtract From Immediate Carrying
XO  31 232 SR  54 subfme[o][.] Subtract From Minus One Extended
XO  31 200 SR  55 subfze[o][.] Subtract From Zero Extended
X   31 598     II sync         Synchronize
X   31  68     64 td           Trap Doubleword
D    2         63 tdi          Trap Doubleword Immediate
X   31 370    III tlbia        TLB Invalidate All
X   31 306 64 III tlbie        TLB Invalidate Entry
X   31 566    III tlbsync      TLB Synchronize
X   31  4      64 tw           Trap Word
D    3         63 twi          Trap Word Immediate
X   31 316 SR  67 xor[.]       XOR
D   26         66 xori         XOR Immediate
D   27         66 xoris        XOR Immediate Shifted
EOL
		map.each { |l|
			raise 'PPC opcode list error' unless l =~ /^(...) (..) (...) (..) ... (.*)/
			type, opnum, opxnum, mode, name = [$1, $2, $3, $4, $5].map { |w| w.strip }
			opnum, opxnum = opnum.to_i, opxnum.to_i

			pre_addop type, opnum, opxnum, mode, name
		}
	end

	# call addop_X(..), handles name[o][.]
	def pre_addop(type, opnum, opxnum, mode, name)
		if name.include?('[o]')
			pre_addop(type, opnum, opxnum, mode, name.sub('[o]', ''))
			pre_addop(type, opnum, opxnum, mode, name.sub('[o]', 'o'))
		elsif name.include?('[.]')
			pre_addop(type, opnum, opxnum, mode, name.sub('[.]', ''))
			pre_addop(type, opnum, opxnum, mode, name.sub('[.]', '.'))
		else
			send "addop_#{type}", opnum, opxnum, mode, name
		end
	end

	def addop_SC(opnum, opxnum, mode, name)
		bin = opnum << 26
		bin |= 1 << 1
		addop name, bin
		addop name, bin, :lev
	end

	def addop_
	end

	def addop(name, bin, *args)
		o = Opcode.new name

		o.bin = bin
		o.args.concat(args & @fields_mask.keys)
		(args & @valid_props).each { |p| o.props[p] = true }

		@opcode_list << o
	end

	def init_mips32_reserved
		addop 'future111011', 0b111011 << 26, :i26

		%w[011000 011001 011010 011011  100111 101100 101101 110100 110111 111100 111111].each { |b|
			addop "reserved#{b}", b.to_i(2) << 26, :i26
		}

		addop 'ase_jalx', 0b011101 << 26, :i26
		addop 'ase011110', 0b011110 << 26, :i26
		# TODO add all special/regimm/... 
	end

	def init_mips32
		@fields_mask.update :rs => 0x1f, :rt => 0x1f, :rd => 0x1f, :sa => 0x1f,
			:i16 => 0xffff, :i26 => 0x3ffffff, :rs_i16 => 0x3e0ffff, :it => 0x1f,
			:ft => 0x1f, :idm1 => 0x1f, :idb => 0x1f, :sel => 7, :i20 => 0xfffff #, :i32 => 0
		@fields_shift.update :rs => 21, :rt => 16, :rd => 11, :sa => 6,
			:i16 => 0, :i26 => 0, :rs_i16 => 0, :it => 16,
			:ft => 16, :idm1 => 11, :idb => 11, :sel => 0, :i20 => 6 #, :i32 => 0

		init_mips32_obsolete
		init_mips32_reserved

		addop 'j',    0b000010 << 26, :i26, :setip, :stopexec	# sets the program counter to (i26 << 2) | ((pc+4) & 0xfc000000) ie i26*4 in the 256M-aligned section containing the instruction in the delay slot
		addop 'jal',  0b000011 << 26, :i26, :setip, :stopexec, :saveip	# same thing, saves return addr in r31

		addop 'mov',  0b001000 << 26, :rt, :rs			# rt <- rs+0
		addop 'addi', 0b001000 << 26, :rt, :rs, :i16		# add		rt <- rs+i
		addop 'addiu',0b001001 << 26, :rt, :rs, :i16		# add unsigned
		addop 'slti', 0b001010 << 26, :rt, :rs, :i16		# set on less than
		addop 'sltiu',0b001011 << 26, :rt, :rs, :i16		# set on less than unsigned
		addop 'andi', 0b001100 << 26, :rt, :rs, :i16		# and
		addop 'ori',  0b001101 << 26, :rt, :rs, :i16		# or
		addop 'xori', 0b001110 << 26, :rt, :rs, :i16		# xor
		addop 'lui',  0b001111 << 26, :rt, :i16			# load upper
#		addop 'li',   0b001111 << 26, :rt, :i32			# pseudoinstruction

		addop 'beq',  0b000100 << 26, :rt, :rs, :i16, :setip	# ==
		addop 'bne',  0b000101 << 26, :rt, :rs, :i16, :setip	# !=
		addop 'blez', 0b000110 << 26, :rs, :i16, :setip		# <= 0
		addop 'bgtz', 0b000111 << 26, :rs, :i16, :setip		# > 0

		addop 'lb',   0b100000 << 26, :rt, :rs_i16		# load byte	rs <- [rt+i]
		addop 'lh',   0b100001 << 26, :rt, :rs_i16		# load halfword
		addop 'lwl',  0b100010 << 26, :rt, :rs_i16		# load word left
		addop 'lw',   0b100011 << 26, :rt, :rs_i16		# load word
		addop 'lbu',  0b100100 << 26, :rt, :rs_i16		# load byte unsigned
		addop 'lhu',  0b100101 << 26, :rt, :rs_i16		# load halfword unsigned
		addop 'lwr',  0b100110 << 26, :rt, :rs_i16		# load word right

		addop 'sb',   0b101000 << 26, :rt, :rs_i16		# store byte
		addop 'sh',   0b101001 << 26, :rt, :rs_i16		# store halfword
		addop 'swl',  0b101010 << 26, :rt, :rs_i16		# store word left
		addop 'sw',   0b101011 << 26, :rt, :rs_i16		# store word
		addop 'swr',  0b101110 << 26, :rt, :rs_i16		# store word right

		addop 'll',   0b110000 << 26, :rt, :rs_i16		# load linked word (read for atomic r/modify/w, sc does the w)
		addop 'sc',   0b111000 << 26, :rt, :rs_i16		# store conditional word

		addop 'lwc1', 0b110001 << 26, :ft, :rs_i16		# load word in fpreg low
		addop 'swc1', 0b111001 << 26, :ft, :rs_i16		# store low fpreg word
		addop 'lwc2', 0b110010 << 26, :rt, :rs_i16		# load word to copro2 register low
		addop 'swc2', 0b111010 << 26, :rt, :rs_i16		# store low coproc2 register

		addop 'ldc1', 0b110101 << 26, :ft, :rs_i16		# load dword in fpreg low
		addop 'sdc1', 0b111101 << 26, :ft, :rs_i16		# store fpreg
		addop 'ldc2', 0b110110 << 26, :rt, :rs_i16		# load dword to copro2 register
		addop 'sdc2', 0b111110 << 26, :rt, :rs_i16		# store coproc2 register

		addop 'pref', 0b110011 << 26, :it, :rs_i16		# prefetch (it = %w[load store r2 r3 load_streamed store_streamed load_retained store_retained
									# r8 r9 r10 r11 r12 r13 r14 r15 r16 r17 r18 r19 r20 r21 r22 r23 r24 writeback_invalidate
									# id26 id27 id28 id29 prepare_for_store id31]
		addop 'cache',0b101111 << 26, :it, :rs_i16		# do things with the proc cache

		# special
		addop 'nop',  0
		addop 'ssnop',1<<6
		addop 'ehb',  3<<6
		addop 'sll',  0b000000, :rd, :rt, :sa
		addop 'movf', 0b000001, :rd, :rs, :cc
		addop 'movt', 0b000001 | (1<<16), :rd, :rs, :cc
		addop 'srl',  0b000010, :rd, :rt, :sa
		addop 'sra',  0b000011, :rd, :rt, :sa
		addop 'sllv', 0b000100, :rd, :rt, :rs
		addop 'srlv', 0b000110, :rd, :rt, :rs
		addop 'srav', 0b000111, :rd, :rt, :rs
		
		addop 'jr',   0b001000, :rs, :setip, :stopexec			# hint field ?
		addop 'jr.hb',0b001000 | (1<<10), :rs, :setip, :stopexec
		addop 'jalr', 0b001001 | (31<<11), :rs, :setip, :stopexec, :saveip	# rd = r31 implicit
		addop 'jalr', 0b001001, :rd, :rs, :setip, :stopexec, :saveip
		addop 'jalr.hb', 0b001001 | (1<<10) | (31<<11), :rs, :setip, :stopexec, :saveip
		addop 'jalr.hb', 0b001001 | (1<<10), :rd, :rs, :setip, :stopexec, :saveip
		addop 'movz', 0b001010, :rd, :rs, :rt			# rt == 0 ? rd <- rs
		addop 'movn', 0b001011, :rd, :rs, :rt
		addop 'syscall', 0b001100, :i20
		addop 'break',0b001101, :i20
		addop 'sync', 0b001111					# type 0 implicit
		addop 'sync', 0b001111, :sa

		addop 'mfhi', 0b010000, :rd				# copies special reg HI to reg
		addop 'mthi', 0b010001, :rd				# copies reg to special reg HI
		addop 'mflo', 0b010010, :rd				# copies special reg LO to reg
		addop 'mtlo', 0b010011, :rd				# copies reg to special reg LO

		addop 'mult', 0b011000, :rs, :rt			# multiplies the registers and store the result in HI:LO
		addop 'multu',0b011001, :rs, :rt
		addop 'div',  0b011010, :rs, :rt
		addop 'divu', 0b011011, :rs, :rt
		addop 'add',  0b100000, :rd, :rs, :rt
		addop 'addu', 0b100001, :rd, :rs, :rt
		addop 'sub',  0b100010, :rd, :rs, :rt
		addop 'subu', 0b100011, :rd, :rs, :rt
		addop 'and',  0b100100, :rd, :rs, :rt
		addop 'or',   0b100101, :rd, :rs, :rt
		addop 'xor',  0b100110, :rd, :rs, :rt
		addop 'nor',  0b100111, :rd, :rs, :rt
	
		addop 'slt',  0b101010, :rd, :rs, :rt			# rs<rt ? rd<-1 : rd<-0
		addop 'sltu', 0b101011, :rd, :rs, :rt

		addop 'tge',  0b110000, :rs, :rt			# rs >= rt ? trap
		addop 'tgeu', 0b110001, :rs, :rt
		addop 'tlt',  0b110010, :rs, :rt
		addop 'tltu', 0b110011, :rs, :rt
		addop 'teq',  0b110100, :rs, :rt
		addop 'tne',  0b110110, :rs, :rt


		# regimm
		addop 'bltz', (1<<26) | (0b00000<<16), :rs, :i16, :setip
		addop 'bgez', (1<<26) | (0b00001<<16), :rs, :i16, :setip
		addop 'tgei', (1<<26) | (0b01000<<16), :rs, :i16, :setip
		addop 'tgfiu',(1<<26) | (0b01001<<16), :rs, :i16, :setip
		addop 'tlti', (1<<26) | (0b01010<<16), :rs, :i16, :setip
		addop 'tltiu',(1<<26) | (0b01011<<16), :rs, :i16, :setip
		addop 'teqi', (1<<26) | (0b01100<<16), :rs, :i16, :setip
		addop 'tnei', (1<<26) | (0b01110<<16), :rs, :i16, :setip
		addop 'bltzal', (1<<26) | (0b10000<<16), :rs, :i16, :setip
		addop 'bgezal', (1<<26) | (0b10001<<16), :rs, :i16, :setip


		# special2
		addop 'madd', (0b011100<<26) | 0b000000, :rs, :rt
		addop 'maddu',(0b011100<<26) | 0b000001, :rs, :rt
		addop 'mul',  (0b011100<<26) | 0b000010, :rd, :rs, :rt
		addop 'msub', (0b011100<<26) | 0b000100, :rs, :rt
		addop 'msubu',(0b011100<<26) | 0b000101, :rs, :rt
		addop 'clz',  (0b011100<<26) | 0b100000, :rd, :rs, :rt	# must have rs == rt
		addop 'clo',  (0b011100<<26) | 0b100001, :rd, :rs, :rt	# must have rs == rt
		addop 'sdbbp',(0b011100<<26) | 0b111111, :i20


		# cp0
		addop 'mfc0', (0b010000<<26) | (0b00000<<21), :rt, :rd
		addop 'mfc0', (0b010000<<26) | (0b00000<<21), :rt, :rd, :sel
		addop 'mtc0', (0b010000<<26) | (0b00100<<21), :rt, :rd
		addop 'mtc0', (0b010000<<26) | (0b00100<<21), :rt, :rd, :sel

		addop 'tlbr', (0b010000<<26) | (1<<25) | 0b000001
		addop 'tlbwi',(0b010000<<26) | (1<<25) | 0b000010
		addop 'tlbwr',(0b010000<<26) | (1<<25) | 0b000110
		addop 'tlbp', (0b010000<<26) | (1<<25) | 0b001000
		addop 'eret', (0b010000<<26) | (1<<25) | 0b011000
		addop 'deret',(0b010000<<26) | (1<<25) | 0b011111
		addop 'wait', (0b010000<<26) | (1<<25) | 0b100000	# mode field ?
	end

	def init_mips32r2
		init_mips32

		addop 'rotr', 0b000010 | (1<<21), :rd, :rt, :sa
		addop 'rotrv',0b000110 | (1<<6), :rd, :rt, :rs

		addop 'synci',(1<<26) | (0b11111<<16), :rs_i16

		# special3
		addop 'ext', (0b011111<<26) | 0b000000, :rt, :rs, :sa, :idm1
		addop 'ins', (0b011111<<26) | 0b000100, :rt, :rs, :sa, :idb
		addop 'rdhwr',(0b011111<<26)| 0b111011, :rt, :rd
		addop 'wsbh',(0b011111<<26) | (0b00010<<6) | 0b100000, :rd, :rt
		addop 'seb', (0b011111<<26) | (0b10000<<6) | 0b100000, :rd, :rt
		addop 'seh', (0b011111<<26) | (0b11000<<6) | 0b100000, :rd, :rt

		# cp0
		addop 'rdpgpr', (0b010000<<26) | (0b01010<<21), :rd, :rt
		addop 'wrpgpr', (0b010000<<26) | (0b01110<<21), :rd, :rt
		addop 'di',     (0b010000<<26) | (0b01011<<21) | (0b01100<<11) | (0<<5)
		addop 'di',     (0b010000<<26) | (0b01011<<21) | (0b01100<<11) | (0<<5), :rt
		addop 'ei',     (0b010000<<26) | (0b01011<<21) | (0b01100<<11) | (1<<5)
		addop 'ei',     (0b010000<<26) | (0b01011<<21) | (0b01100<<11) | (1<<5), :rt
	end
end
end
