require 'metasm'
include Metasm
require 'optparse'

# usage: patchelf.rb targetelf [-o outputelf] [--c] addrtopatch srcofpatch addrtopatch src..
outfile = nil
c_file = false
OptionParser.new { |opt|
	opt.on('-o outfile') { |f| outfile = f }
	opt.on('--c') { c_file = true }
}.parse!(ARGV)

target = ARGV.shift
outfile ||= target + '.patched'

e = ELF.decode_file target
cpu = e.cpu_from_headers

patches = {}
while patchaddr = ARGV.shift
	patchaddr = Integer(patchaddr) if (?0..?9).include? patchaddr[0]
	if not patchaddr.kind_of? Integer
		s = e.segments.find { |s| s.encoded and s.encoded.export[patchaddr] }
		patchaddr = s.vaddr + s.encoded.export[patchaddr] if s
		raise "invalid address #{patchaddr}" if not patchaddr.kind_of? Integer
	end

	hook_label = elf.new_label('hook')
	hook = Shellcode.assemble(cpu, "jmp #{hook_label}").encoded
	e.encoded.ptr = e.addr_to_off(patchaddr)
	patches[e.encoded.ptr] = hook
	overwritten, osz = [], 0
	while osz < hook.length
		di = cpu.decode_instruction(e.encoded, patchaddr+osz)
		raise "unknown instruction overwritten by hook for #{Expression[patchaddr]}" if not di.kind_of? DecodedInstruction
		overwritten << di.instruction
		osz += di.bin_length
	end
	patchasm = "#{hook_label}:\n"
	patchasm << <<EOS
pushad pushfd
lea eax, [esp+8*4]
push eax
call 1f
add esp, 4
popfd popad
#{overwritten.join("\n")}
jmp #{patchaddr}+#{osz}
1:
EOS
	patchasm << File.read(ARGV.shift)	# must be a __cdecl void hook(void* orig_ctx);
	patchcode = Shellcode.assemble(cpu, patchasm).encoded
end

patchseg = ELF::Segment.new
patchseg.type = 'LOAD'
patchseg.flags = ['R', 'X']
patchseg.encoded = EncodedData.new << patchcode

e.segments << s
s = ELF::Segment.new
s.type = 'LOAD'
s.flags = ['R', 'W']
s.encoded = EncodedData.new
e.segments << s

new_elf = e.encoded.dup
e.header.phoff = new_elf.length
new_hdr = e.header.encode(e)
new_elf[0, new_hdr.length] = new_hdr
e.segments.each { |s| new_elf << s.encode(e) }

binding = e.symbols.inject({}).map { |h, s| next h if not s.binding == 'GLOBAL' ; h.update s.name => s.value }
binding.update s.encoded.

new_elf << e.segments[-2].encoded
new_elf << e.segments[-1].encoded
new_elf.fixup(binding)


